typedef struct cuda_id_struct
{
    int x;
    int y;
} cuda_id;

__kernel void kernel_conv_global(__global unsigned char* gInput,
                                 __global unsigned char* gOutput,
                                 __constant int *filter_coeffs,
								 int imgWidth,
								 int imgWidthF)
{
  cuda_id threadIdx, blockIdx, blockDim;
  threadIdx.x = get_local_id(0);
  threadIdx.y = get_local_id(1);
  blockIdx.x  = get_group_id(0);
  blockIdx.y  = get_group_id(1);
  blockDim.x  = get_local_size(0);
  blockDim.y  = get_local_size(1);
  
  
  int row = blockIdx.y * blockDim.y + threadIdx.y;
  int col = blockIdx.x * blockDim.x + threadIdx.x;
  
  
  int pix_data[3] = {0, 0, 0};

  int in_pix = (row*imgWidthF + col) * 3;;
  #pragma unroll 5
  for (int fy=0; fy<5; fy++)
  {
      #pragma unroll 5
      for (int fx=0; fx<5; fx++)
      {
          #pragma unroll
          for (int rgba=0; rgba<3; rgba++)
          {
              pix_data[rgba] = pix_data[rgba] + (gInput[in_pix+rgba] * filter_coeffs[fy*5+fx]);
          }
          in_pix=in_pix+3;
      }
      in_pix = in_pix + 3*imgWidthF - 5*1*3;
  }

  int out_pix = (row*imgWidth + col) * 3;
  for (int rgba=0; rgba<3; rgba++)
  {
      pix_data[rgba] = min(pix_data[rgba], 255);
      pix_data[rgba] = max(pix_data[rgba],   0);
      gOutput[out_pix+rgba] = (unsigned char)(pix_data[rgba]);
  }
}


//----------------------------------------------------------------------------------------------------------------------------


__kernel void kernel_conv_sh_uchar_int(__global unsigned char* gInput,
                                       __global unsigned char* gOutput,
                                       __constant int *filter_coeffs,
									   int imgWidth,
                                       int imgWidthF)
{
  cuda_id threadIdx, blockIdx, blockDim;
  threadIdx.x = get_local_id(0);
  threadIdx.y = get_local_id(1);
  blockIdx.x  = get_group_id(0);
  blockIdx.y  = get_group_id(1);
  blockDim.x  = get_local_size(0);
  blockDim.y  = get_local_size(1);

  int row = blockIdx.y * blockDim.y + threadIdx.y;
  int col = blockIdx.x * blockDim.x + threadIdx.x;
  
  int pix_data[3] = {0, 0, 0};

  __local unsigned char in_shmem[20][20][3];

  int thid_1d = threadIdx.y*blockDim.x + threadIdx.x;
  int ld_rgba = thid_1d%3;
  int ld_col  = (thid_1d/3)%20;
  int ld_row  = thid_1d / 60;
  int ld_base = (blockIdx.y * blockDim.y) * 3 * imgWidthF + (blockIdx.x * blockDim.x) * 3 + (ld_row * 3 * imgWidthF);
  if (thid_1d<(3*20*4))
  {
	  #pragma unroll
	  for (int ld=0; ld<5; ld++)
	  {
		  in_shmem[ld_row + ld * 4][ld_col][ld_rgba] = gInput[ld_base + (thid_1d % 60)];
		  ld_base = ld_base + imgWidthF*3*4;
	  }
  }
  barrier(CLK_LOCAL_MEM_FENCE);

  #pragma unroll 5
  for (int fy=0; fy<5; fy++)
  {
      #pragma unroll 5
      for (int fx=0; fx<5; fx++)
      {
		  int coeff =  filter_coeffs[fy*5+fx];
          #pragma unroll 3
          for (int rgba=0; rgba<3; rgba++)
          {
			  int pix = (int)(in_shmem[threadIdx.y+fy][threadIdx.x+fx][rgba]);
              pix_data[rgba] = pix_data[rgba] + (coeff * pix);
          }
      }
  }

  int out_pix = (row*imgWidth + col) * 3;
  for (int rgba=0; rgba<3; rgba++)
  {
      pix_data[rgba] = min(pix_data[rgba], 255);
      pix_data[rgba] = max(pix_data[rgba],   0);
	  gOutput[out_pix + rgba] = (unsigned char)(pix_data[rgba]);
  }
}



//----------------------------------------------------------------------------------------------------------------------------
__kernel void kernel_conv_sh_uchar_float(__global unsigned char* gInput,
                                         __global unsigned char* gOutput,
                                         __constant float *filter_coeffs,
										 int imgWidth,
                                         int imgWidthF)
{
  cuda_id threadIdx, blockIdx, blockDim;
  threadIdx.x = get_local_id(0);
  threadIdx.y = get_local_id(1);
  blockIdx.x  = get_group_id(0);
  blockIdx.y  = get_group_id(1);
  blockDim.x  = get_local_size(0);
  blockDim.y  = get_local_size(1);

	int row = blockIdx.y * blockDim.y + threadIdx.y;
	int col = blockIdx.x * blockDim.x + threadIdx.x;

	
	float pix_data[3] = { 0.0f, 0.0f, 0.0f };

	__local unsigned char in_shmem[20][20][3];

	int thid_1d = threadIdx.y*blockDim.x + threadIdx.x;
	int ld_row = thid_1d / 60;
	int ld_addr = (blockIdx.y * blockDim.y) * 3 * imgWidthF + (blockIdx.x * blockDim.x) * 3 + (ld_row * 3 * imgWidthF) + (thid_1d % 60);
	int smem_addr = thid_1d;
	__local unsigned char *smem_base_addr = &in_shmem[0][0][0];
	if (thid_1d<3 * 20 * 4)
	{
#pragma unroll
		for (int ld = 0; ld<5; ld++)
		{
			
			*(smem_base_addr + smem_addr) = gInput[ld_addr];
			ld_addr = ld_addr + imgWidthF * 3 * 4;
			smem_addr = smem_addr + 4 * 60;
		}
	}
	barrier(CLK_LOCAL_MEM_FENCE);

#pragma unroll 5
	for (int fy = 0; fy<5; fy++)
	{
#pragma unroll 5
		for (int fx = 0; fx<5; fx++)
		{
			float coeff = filter_coeffs[fy*5+fx];
#pragma unroll 3
			for (int rgba = 0; rgba<3; rgba++)
			{
				float pix = (float)(in_shmem[threadIdx.y + fy][threadIdx.x + fx][rgba]);
				pix_data[rgba] = pix_data[rgba] + (coeff * pix);
			}
		}
	}

	int out_pix = (row*imgWidth + col) * 3;
	for (int rgba = 0; rgba<3; rgba++)
	{
		pix_data[rgba] = min(pix_data[rgba], 255.0f);
		pix_data[rgba] = max(pix_data[rgba], 0.0f);
		gOutput[out_pix + rgba] = (unsigned char)(pix_data[rgba]);
	}
}



//----------------------------------------------------------------------------------------------------------------------------
__kernel void kernel_conv_sh_float_float(__global unsigned char* gInput,
                                         __global unsigned char* gOutput,
                                         __constant float *filter_coeffs,
										 int imgWidth,
                                         int imgWidthF)
{
  cuda_id threadIdx, blockIdx, blockDim;
  threadIdx.x = get_local_id(0);
  threadIdx.y = get_local_id(1);
  blockIdx.x  = get_group_id(0);
  blockIdx.y  = get_group_id(1);
  blockDim.x  = get_local_size(0);
  blockDim.y  = get_local_size(1);
  
  
	int row = blockIdx.y * blockDim.y + threadIdx.y;
	int col = blockIdx.x * blockDim.x + threadIdx.x;

	
	float pix_data[3] = { 0.0f, 0.0f, 0.0f };

	__local float in_shmem[20][20][3];

	int thid_1d = threadIdx.y*blockDim.x + threadIdx.x;
	int ld_rgba = thid_1d % 3;
	int ld_col = (thid_1d / 3) % 20;
	int ld_row = thid_1d / 60;
	int ld_base = (blockIdx.y * blockDim.y) * 3 * imgWidthF + (blockIdx.x * blockDim.x) * 3 + (ld_row * 3 * imgWidthF);
	if (thid_1d<3 * 20 * 4)
	{
#pragma unroll
		for (int ld = 0; ld<5; ld++)
		{
			in_shmem[ld_row + ld * 4][ld_col][ld_rgba] = (float)(gInput[ld_base + (thid_1d % 60)]);
			ld_base = ld_base + imgWidthF * 3 * 4;
		}
	}
	barrier(CLK_LOCAL_MEM_FENCE);

#pragma unroll 5
	for (int fy = 0; fy<5; fy++)
	{
#pragma unroll 5
		for (int fx = 0; fx<5; fx++)
		{
			float coeff = filter_coeffs[fy*5+fx];
#pragma unroll
			for (int rgba = 0; rgba<3; rgba++)
			{
				float pix = in_shmem[threadIdx.y + fy][threadIdx.x + fx][rgba];
				pix_data[rgba] = pix_data[rgba] + (coeff * pix);
			}
		}
	}

	int out_pix = (row*imgWidth + col) * 3;
	for (int rgba = 0; rgba<3; rgba++)
	{
		pix_data[rgba] = min(pix_data[rgba], 255.0f);
		pix_data[rgba] = max(pix_data[rgba], 0.0f);
		gOutput[out_pix + rgba] = (unsigned char)(pix_data[rgba]);
	}
}



//----------------------------------------------------------------------------------------------------------------------------
__kernel void kernel_conv_sh_float_float_nbc(__global unsigned char* gInput,
                                             __global unsigned char* gOutput,
                                             __constant float *filter_coeffs,
											 int imgWidth,
                                             int imgWidthF)
{
  cuda_id threadIdx, blockIdx, blockDim;
  threadIdx.x = get_local_id(0);
  threadIdx.y = get_local_id(1);
  blockIdx.x  = get_group_id(0);
  blockIdx.y  = get_group_id(1);
  blockDim.x  = get_local_size(0);
  blockDim.y  = get_local_size(1);
	int row = blockIdx.y * blockDim.y + threadIdx.y;
	int col = blockIdx.x * blockDim.x + threadIdx.x;

	float pix_data[3] = { 0.0f, 0.0f, 0.0f };

	__local float in_shmem[8+4][32+4][3];

	int thid_1d = threadIdx.y*blockDim.x + threadIdx.x;
	int ld_rgba = thid_1d % 3;
	int ld_col = (thid_1d / 3) % (32+4);
	int ld_row = thid_1d / (3*(32+4));
	int ld_base = (blockIdx.y * blockDim.y) * 3 * imgWidthF + (blockIdx.x * blockDim.x) * 3 + (ld_row * 3 * imgWidthF);
	if (thid_1d<(2*3*(32+4)))
	{
#pragma unroll
		for (int ld = 0; ld<6; ld++)
		{
			in_shmem[ld_row + ld * 2][ld_col][ld_rgba] = (float)(gInput[ld_base + (thid_1d % (3 * (32 + 4)))]);
			ld_base = ld_base + imgWidthF * 3 * 2;
		}
	}
	barrier(CLK_LOCAL_MEM_FENCE);

#pragma unroll 5
	for (int fy = 0; fy<5; fy++)
	{
#pragma unroll 5
		for (int fx = 0; fx<5; fx++)
		{
			float coeff = filter_coeffs[fy*5+fx];
#pragma unroll
			for (int rgba = 0; rgba<3; rgba++)
			{
				float pix = in_shmem[threadIdx.y + fy][threadIdx.x + fx][rgba];
				pix_data[rgba] = pix_data[rgba] + (coeff * pix);
			}
		}
	}

	int out_pix = (row*imgWidth + col) * 3;
	for (int rgba = 0; rgba<3; rgba++)
	{
		pix_data[rgba] = min(pix_data[rgba], 255.0f);
		pix_data[rgba] = max(pix_data[rgba], 0.0f);
		gOutput[out_pix + rgba] = (unsigned char)(pix_data[rgba]);
	}
}
